@@ -289,6 +289,7 @@ class Detectors:
         thi = False
         thf_list = []
         thf = False
+        newM5 = False
 
         for i in range(len(low_pass)):
 
@@ -306,7 +307,7 @@ class Detectors:
                 if newM5>1.5*MM[-1]:
                     newM5 = 1.1*MM[-1]
 
-            elif QRS and i == QRS[-1]+ms200:
+            elif newM5 and QRS and i == QRS[-1]+ms200:
                 MM.append(newM5)
                 if len(MM)>5:
                     MM.pop(0)    
@@ -360,6 +361,8 @@ class Detectors:
                 thi = False
                 thf = False
 
+        # removing the 1st detection as it 1st needs the QRS complex amplitude for the threshold
+        r_peaks.pop(0)
         return r_peaks
 
     
@@ -443,7 +446,7 @@ class Detectors:
         return filt_peaks
 
 
-    def pan_tompkins_detector(self, unfiltered_ecg):
+    def pan_tompkins_detector(self, unfiltered_ecg, MWA_name='cumulative'):
         """
         Jiapu Pan and Willis J. Tompkins.
         A Real-Time QRS Detection Algorithm. 
@@ -463,7 +466,7 @@ class Detectors:
         squared = diff*diff
 
         N = int(0.12*self.fs)
-        mwa = MWA(squared, N)
+        mwa = MWA_from_name(MWA_name)(squared, N)
         mwa[:int(0.2*self.fs)] = 0
 
         mwa_peaks = panPeakDetect(mwa, self.fs)
@@ -471,7 +474,7 @@ class Detectors:
         return mwa_peaks
 
 
-    def two_average_detector(self, unfiltered_ecg):
+    def two_average_detector(self, unfiltered_ecg, MWA_name='cumulative'):
         """
         Elgendi, Mohamed & Jonkman, 
         Mirjam & De Boer, Friso. (2010).
@@ -488,10 +491,10 @@ class Detectors:
         filtered_ecg = signal.lfilter(b, a, unfiltered_ecg)
 
         window1 = int(0.12*self.fs)
-        mwa_qrs = MWA(abs(filtered_ecg), window1)
+        mwa_qrs = MWA_from_name(MWA_name)(abs(filtered_ecg), window1)
 
         window2 = int(0.6*self.fs)
-        mwa_beat = MWA(abs(filtered_ecg), window2)
+        mwa_beat = MWA_from_name(MWA_name)(abs(filtered_ecg), window2)
 
         blocks = np.zeros(len(unfiltered_ecg))
         block_height = np.max(filtered_ecg)
@@ -521,23 +524,96 @@ class Detectors:
 
         return QRS
 
+    def wqrs_detector(self, unfiltered_ecg):
+        """
+        based on W Zong, GB Moody, D Jiang 
+        A Robust Open-source Algorithm to Detect Onset and Duration of QRS
+        Complexes 
+        In: 2003 IEEE
+        """
+        def butter_lowpass_filter(data, cutoff):
+            nyq = 0.5 * self.fs
+            order = 2
 
-def MWA(input_array, window_size):
+            normal_cutoff = cutoff / nyq
+            
+            b, a = signal.butter(order, normal_cutoff, btype='low', analog=False)
+            y = signal.lfilter(b, a, data)
+            return y
+
+        def length_transfrom(x, w):
+            tmp = []
+            for i in range(w, len(x)):
+                curr = 0
+                for k in range(i-w+1, i):
+                    curr += np.sqrt(np.power(1/self.fs, 2)+np.power(x[i]-x[i-1],2))
+                tmp.append(curr)
+            l = [tmp[0]]*w
+            
+            return l+tmp
+        
+        def threshold(x):
+            peaks = []
+            u = MWA_convolve(x, 10*self.fs)
+            for i in range(len(x)):
+                if (len(peaks) == 0 or i > peaks[-1]+(self.fs*0.35)) and x[i] > u[i]:
+                    peaks.append(i)
+            return peaks
+        
+        y = butter_lowpass_filter(unfiltered_ecg, 15)
+        y = length_transfrom(y, int(np.ceil(self.fs*0.13)))
+        return threshold(y)
+
+def MWA_from_name(function_name):
+    if function_name == "cumulative":
+        return MWA_cumulative
+    elif function_name == "convolve":
+        return MWA_convolve
+    elif function_name == "original":
+        return MWA_original
+    else: 
+        raise RuntimeError('invalid moving average function!')
+
+#Fast implementation of moving window average with numpy's cumsum function 
+def MWA_cumulative(input_array, window_size):
+    
+    ret = np.cumsum(input_array, dtype=float)
+    ret[window_size:] = ret[window_size:] - ret[:-window_size]
+    
+    for i in range(1,window_size):
+        ret[i-1] = ret[i-1] / i
+    ret[window_size - 1:]  = ret[window_size - 1:] / window_size
+    
+    return ret
+
+#Original Function 
+def MWA_original(input_array, window_size):
 
     mwa = np.zeros(len(input_array))
-    for i in range(len(input_array)):
+    mwa[0] = input_array[0]
+    
+    for i in range(2,len(input_array)+1):
         if i < window_size:
             section = input_array[0:i]
         else:
-            section = input_array[i-window_size:i]
+            section = input_array[i-window_size:i]        
         
-        if i!=0:
-            mwa[i] = np.mean(section)
-        else:
-            mwa[i] = input_array[i]
+        mwa[i-1] = np.mean(section)
 
     return mwa
 
+#Fast moving window average implemented with 1D convolution 
+def MWA_convolve(input_array, window_size):
+    
+    ret = np.pad(input_array, (window_size-1,0), 'constant', constant_values=(0,0))
+    ret = np.convolve(ret,np.ones(window_size),'valid')
+    
+    for i in range(1,window_size):
+        ret[i-1] = ret[i-1] / i
+    ret[window_size-1:] = ret[window_size-1:] / window_size
+    
+    return ret
+
 
 def normalise(input_array):
 
